import common;
import scene;
import windows;
import std.math;
debug import std.stdio;

struct Ray {
  static Ray opCall( vector3 a_Origin, vector3 a_Dir ) {
    Ray r;
    r.m_Origin = a_Origin;
    r.m_Direction = a_Dir;
    return r;
  }
  vector3 origin;
  vector3 direction;
  
  void SetOrigin( vector3 a_Origin ) {
    m_Origin = a_Origin;
  }
  void SetDirection( vector3 a_Direction ) {
    m_Direction = a_Direction;
  }
  vector3 GetOrigin() {
    return m_Origin;
  }
  vector3 GetDirection() {
    return m_Direction;
  }
 private:
  alias origin m_Origin;
  alias direction m_Direction;
}

class Engine {
  this() { m_Scene = new Scene(); }

  /**
   * Sets the render target canvas
   */
  void SetTarget( Pixel * a_Dest, int a_Width, int a_Height ) {
    // set pixel buffer address & size
    m_Dest = a_Dest;
    m_Width = a_Width;
    m_Height = a_Height;
  }

  Scene GetScene() {
    return m_Scene;
  }

  /**
   * Naive ray tracing: Intersects the ray with every primitive
   * in the scene to determine the closest intersection
   */
  Primitive Raytrace( Ray a_Ray, inout Color a_Acc, int a_Depth, float a_RIndex, inout float a_Dist ) {
    if (a_Depth > TRACEDEPTH) {
      return null;
    }
    // trace primary ray
    a_Dist = 1000000.0f;
    vector3 pi;
    Primitive prim;
    int result;
    // find the nearest intersection
    foreach (Primitive pr; m_Scene) {
      int res;
      if ((res = pr.Intersect( a_Ray, a_Dist )) != 0) {
        prim = pr;
        result = res; // 0 = miss, 1 = hit, -1 = hit from inside primitive
      }
    }
    // no hit, terminate ray
    if (prim is null) {
      return null;
    }
    // handle intersection
    if (prim.IsLight()) {
      // we hit a light, stop tracing
      a_Acc.Set( 1, 1, 1 );
    } else {
      // determine color at point of intersection
      pi = a_Ray.origin + a_Ray.direction * a_Dist;
      // trace lights
      foreach (Primitive p; m_Scene) {
        if (p.IsLight()) {
          Primitive light = p;
          // handle point light source
          float shade = 1.0f;
          if (light.GetType() == Primitive.SPHERE) {
            vector3 L = (cast(Sphere)light).GetCentre() - pi;
            float tdist = LENGTH!(float,vector3)( L );
            L *= (1.0f / tdist);
            Ray r = Ray( pi + L * EPSILON, L );
            foreach (Primitive pr; m_Scene) {
              if ((pr != light) && (pr.Intersect( r, tdist ))) {
                shade = 0;
                break;
              }
            }
          }
          if (shade > 0) {
             // calculate diffuse shading
            vector3 L = (cast(Sphere)light).GetCentre() - pi;
            NORMALIZE( L );
            vector3 N = prim.GetNormal( pi );
            if (prim.GetMaterial().GetDiffuse() > 0) {
              float dot = DOT!(float, vector3) ( N, L );
              if (dot > 0) {
                float diff = dot * prim.GetMaterial().GetDiffuse() * shade;
                // add diffuse component to ray color
                a_Acc += diff * prim.GetMaterial().GetColor() * light.GetMaterial().GetColor();
              }
            }
            // determine specular component
            if (prim.GetMaterial().GetSpecular() > 0) {
              // point light source: sample once for specular highlight
              vector3 V = a_Ray.direction;
              vector3 R = L - 2.0f * DOT!(float, vector3)( L, N ) * N;
              float dot = DOT!(float, vector3)( V, R );
              if (dot > 0) {
                float spec = pow( cast(real)dot, 20 ) * prim.GetMaterial().GetSpecular() * shade;
                // add specular component to ray color
                a_Acc += spec * light.GetMaterial().GetColor();
              }
            }
          }
        }
      }
      // calculate reflection
      float refl = prim.GetMaterial().GetReflection();
      if (refl > 0.0f && (a_Depth < TRACEDEPTH)) {
        vector3 N = prim.GetNormal( pi );
        vector3 R = a_Ray.direction - 2.0f * DOT!(float, vector3)( a_Ray.direction, N ) * N;
        Color rcol = Color( 0, 0, 0 );
        float dist;
        Raytrace( Ray( pi + R * EPSILON, R ), rcol, a_Depth + 1, a_RIndex, dist );
        a_Acc += refl * rcol * prim.GetMaterial().GetColor();
      }
      // calculate refraction
      float refr = prim.GetMaterial().GetRefraction();
      if ((refr > 0) && (a_Depth < TRACEDEPTH))
      {
        float rindex = prim.GetMaterial().GetRefrIndex();
        float n = a_RIndex / rindex;
        vector3 N = prim.GetNormal( pi ) * cast(float)result;
        float cosI = -DOT!(float, vector3)( N, a_Ray.direction );
        float cosT2 = 1.0f - n * n * (1.0f - cosI * cosI);
        if (cosT2 > 0.0f)
        {
          vector3 T = (n * a_Ray.direction) + (n * cosI - sqrt( cosT2 )) * N;
          Color rcol = Color( 0, 0, 0 );
          float dist;
          Raytrace( Ray( pi + T * EPSILON, T ), rcol, a_Depth + 1, rindex, dist );
          // apply Beer's law
          Color absorbance = prim.GetMaterial().GetColor() * 0.15f * -dist;
          Color transparency = Color( exp( absorbance.r ), exp( absorbance.g ), exp( absorbance.b ) );
          a_Acc += rcol * transparency;
        }
      }
    }
    // return pointer to primitive hit by primary ray
    return prim;
  }

  /**
   * Initializes the renderer, by resetting the line / tile
   * counters and precalculating some values
   */
  void InitRender() {
    // set firts line to draw to
    m_CurrLine = 20;
    // set pixel buffer address of first pixel
    m_PPos = m_CurrLine * m_Width;
    // screen plane in world space coordinates
    m_WX1 = -4, m_WX2 = 4, m_WY1 = m_SY = 3, m_WY2 = -3;
    // calculate deltas for interpolation
    m_DX = (m_WX2 - m_WX1) / m_Width;
    m_DY = (m_WY2 - m_WY1) / m_Height;
    m_SY += m_CurrLine * m_DY;
    // allocate space to store pointers to primitives for previous line
    m_LastRow.length = m_Width;
  }

  /**
   * Fires rays in the scene one scanline at a time, from left
   * to right
   */
  bool Render() {
    // render scene
    static vector3 o = { 0, 0, -5 };
    // initialize timer
    int msecs = GetTickCount();
    // reset last found primitive pointer
    Primitive lastprim;

    // render remaining lines
    for ( int y = m_CurrLine; y < (m_Height - 20); y++ ) {
      m_SX = m_WX1;
      // render pixels for current line
      for ( int x = 0; x < m_Width; x++ ) {
        // fire primary rays
        Color acc = Color(0, 0, 0);
        vector3 dir = vector3( m_SX, m_SY, 0 ) - o;
        NORMALIZE( dir );
        Ray r = Ray( o, dir );
        float dist;
        Primitive prim = Raytrace( r, acc, 1, 1.0f, dist );
        int red, green, blue;
        if (prim != lastprim) {
          lastprim = prim;
          acc.Set(0, 0, 0);
          for ( int tx = -1; tx < 2; tx++ ) {
            for ( int ty = -1; ty < 2; ty++ ) {
              dir = vector3( m_SX + m_DX * tx / 2.0f, m_SY + m_DY * ty / 2.0f, 0 ) - o;
              NORMALIZE( dir );
              r = Ray( o, dir );
              dist = 0;
              prim = Raytrace( r, acc, 1, 1.0f, dist );
            }
          }
          red = cast(int)(acc.r * (256 / 9));
          green = cast(int)(acc.g * (256 / 9));
          blue = cast(int)(acc.b * (256 / 9));
        } else {
          red = cast(int)(acc.r * 256);
          green = cast(int)(acc.g * 256);
          blue = cast(int)(acc.b * 256);
        }
        if (red > 255) {
          red = 255;
        }
        if (green > 255) {
          green = 255;
        }
        if (blue > 255) {
          blue = 255;
        }
        m_Dest[m_PPos++] = (red << 16) + (green << 8) + blue;
        m_SX += m_DX;
      }
      m_SY += m_DY;
      // see if we've been working to long already
      if ((GetTickCount() - msecs) > 100) {
        // return control to windows so the screen gets updated
        m_CurrLine = y + 1;
        return false;
      }
    }
    // all done
    return true;
  }


 protected:
  // renderer data
  float m_WX1, m_WY1, m_WX2, m_WY2, m_DX, m_DY, m_SX, m_SY;
  Scene m_Scene;
  Pixel * m_Dest;
  int m_Width, m_Height, m_CurrLine, m_PPos;
  Primitive[] m_LastRow;
}



